# mHPoT trainer
# 
# training on natural image patches using Fast Persistent Contrastive 
# Divergence or Contrastive Divergence. Samples are drawn from the model
# using Hamiltonian Monte Carlo.
# Input patches are whitened (i.e. re-scaled PCA coefficients of leading 
# eigenvectors).
# The whitening transform is: S^-.5 * U' where U collects in its columns 
# the leading eigenvectors, and S is a diagonal matrix with the corresponsding 
# eigenvalues.
# 
# Free energy: 
# F(x) = sum_j gamma log(1 + .5 sum_f P_fj (sum_i C_fi x_i)^2) +
#        + .5 sum_i x_i^2 - sum_k log(1+exp(sum_i W_ki x_i +b_k)) - sum_i b^v_i x_i
# Probability distribution: p(x) proportional to exp(-F(x))
# 
# This corresponds to an energy function defined over two sets of latent variables.
# The latent variables associated to the first term (the covariance latent variables) are
# gamma distributed, while those associated with the last term (the mean latent variables)
# are Bernoulli distributed.
#
# In the code below, we call:
#  gamma -> coeff
#  P -> FH
#  C -> VF
#  W -> w_mean
#  b -> bias_mean
#  b^v -> bias_vis
#
# For more information aobut the model, see eq. 1 through 6 in:
# M. Ranzato, V. Mnih, G. Hinton, "Generating More Realistic Images Using Gated MRF's". 
# Advances in Neural Information Processing Systems 23 (NIPS 2010), Vancouver, 2010
# available at http://www.cs.toronto.edu/~ranzato/publications/ranzato_nips2010.pdf
#
# Marc'Aurelio Ranzato
# 17 December 2010
# ranzato@cs.toronto.edu

import sys
from pylab import *
import numpy as np
import gnumpy as gpu
from scipy.io import loadmat, savemat
from ConfigParser import *

def compute_energy_mHPoT(data,vel,energy,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,tt0,tt1,tt2,feat,featsq,feat_mean):
    ''' This function computes the value of the hamiltonian at each data point in the batch
    The output is a vector of scalars (as many as data points in the batch)
    Return values: F(x) + .5 vel^2 for each data point.'''
    # potential
    # covariance contribution    
    feat = gpu.dot(VF.T, data)
    featsq = feat**2
    t1 = (1 + 0.5 * gpu.dot(FH.T, featsq)).log() * coeff
    energy = t1.sum(axis = 0)    
    # mean contribution
    feat_mean = gpu.dot(w_mean.T, data) + bias_mean
    tt1 = feat_mean > tt0
    tt2 = 1. - tt1
    tt1 *= feat_mean
    feat_mean = - ((1. + (feat_mean * tt2).exp()).log() + tt1)
    energy += feat_mean.sum(axis = 0)
    # add quadratic term: .5 x^2
    energy += (.5 * data**2).sum(axis = 0)
    # visible bias term: - b x
    energy -= (data * bias_vis).sum(axis = 0)
    # kinetic: .5 vel^2
    energy += (.5 * vel**2).sum(axis = 0)
    return energy

def compute_gradient_mHPoT(data,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,feat,featsq,feat_mean,gradient):
    '''  This function computes the derivatives of F(x) w.r.t. the input (for each input data point)'''    
    # compute derivatives through covariance term in the energy function
    feat = gpu.dot(VF.T, data)
    featsq = feat**2
    t1 = 1 + 0.5 * gpu.dot(FH.T, featsq)
    t2 = coeff / t1
    gradient = gpu.dot(VF, gpu.dot(FH, t2) * feat)
    # add visible bias term
    gradient -= bias_vis
    # add MEAN contribution to gradient
    feat_mean = - (gpu.dot(w_mean.T, data) + bias_mean).logistic()
    gradient += gpu.dot(w_mean, feat_mean) # add - W h
    # add derivative of quadratic term: x
    gradient += data
    return gradient

def draw_HMC_samples(data,negdata,vel,gradient,new_energy,old_energy,VF,FH,coeff,bias_vis,w_mean,bias_mean,hmc_step,hmc_step_nr,hmc_ave_rej,hmc_target_ave_rej,t1,t2,t4,t5,tt0,tt1,tt2,thresh,feat,featsq,batch_size,feat_mean):
    ''' Hybrid Monte Carlo sampler
    This is used to draw samples from the model (defined in terms of an energy function over the input pixels)'''
    vel = vel.rand('normal')
    negdata = data.copy()
    old_energy = compute_energy_mHPoT(negdata,vel,old_energy,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,tt0,tt1,tt2,feat,featsq,feat_mean)
    gradient = compute_gradient_mHPoT(negdata,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,feat,featsq,feat_mean,gradient)
    # half step
    vel -= 0.5*hmc_step * gradient
    negdata += hmc_step * vel
    # full leap-frog steps
    for ss in range(hmc_step_nr - 1):
        ## re-evaluate the gradient
        gradient = compute_gradient_mHPoT(negdata,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,feat,featsq,feat_mean,gradient)
        # update variables
        vel -= hmc_step * gradient 
        negdata += hmc_step * vel
    # final half-step
    gradient = compute_gradient_mHPoT(negdata,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,feat,featsq,feat_mean,gradient)
    vel -= 0.5 * hmc_step * gradient
    # compute new energy
    new_energy = compute_energy_mHPoT(negdata,vel,new_energy,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,tt0,tt1,tt2,feat,featsq,feat_mean)
    # rejecton
    thresh = (old_energy - new_energy).exp()
    t4 = t4.rand() > thresh
    # update negdata and rejection rate
    t5 = t4.sum()
    rej = t5/batch_size
    negdata -= negdata*t4 
    negdata += data*t4
    hmc_ave_rej = 0.9*hmc_ave_rej + 0.1*rej # running average rejection rate
    # adjust step size to maintain target rejection rate
    if hmc_ave_rej < hmc_target_ave_rej:
        hmc_step = min(hmc_step*1.01,0.25)
    else:
        hmc_step = max(hmc_step*0.99,.001)
    return negdata, hmc_step, hmc_ave_rej

def train_mHPoT():
    ''' mHPoT trainer: sweeps over the training set to update the parameters.
    For each batch of samples compute derivatives w.r.t. parameters
    at the training samples and at the negative samples drawn by HMC sampler.
    Training using either Contrastive Divergence or Fast Persistent
    Contrastive Divergence.'''

    config = ConfigParser()
    config.read('input_configuration')
    verbose = config.getint('VERBOSITY','verbose')
    num_epochs = config.getint('MAIN_PARAMETER_SETTING','num_epochs') # number of sweeps through training set
    batch_size = config.getint('MAIN_PARAMETER_SETTING','batch_size') # number of samples in each batch
    startFH = config.getint('MAIN_PARAMETER_SETTING','startFH') # number of epochs after which FH matrix is updated
    startwd = config.getint('MAIN_PARAMETER_SETTING','startwd') # number of epochs after which parameters are decayed to 0
    doFPCD = config.getint('MAIN_PARAMETER_SETTING','doFPCD') # flag to decide whether to use Fast PCD or CD1
    apply_mask = 0 	
    # model parameters
    num_fac = config.getint('MODEL_PARAMETER_SETTING','num_fac') # number of filters in VF matrix
    num_hid_cov =  config.getint('MODEL_PARAMETER_SETTING','num_hid_cov') # number of covariance hidden units
    num_hid_mean =  config.getint('MODEL_PARAMETER_SETTING','num_hid_mean') # number of mean hidden units
    # load data
    data_file_name =  config.get('DATA','data_file_name')
    d = loadmat(data_file_name) # input in the format PxD (P vectorized samples with D dimensions)
    pca_file_name =  config.get('DATA','pca_file_name')
    pca = loadmat(pca_file_name) # input in the format VxD
    totnumcases = d["patches"].shape[0] # number of training samples
    d = d["patches"][0:floor(totnumcases/batch_size)*batch_size,:].copy() 
    totnumcases = d.shape[0]
    num_vis =   pca["pcatransf"].shape[0] # dimensionality of the input (after PCA projection)
    num_batches = int(totnumcases/batch_size) # number of batches
    dev_dat = gpu.garray(np.dot(pca["pcatransf"],d.T)) # VxP, P is number of samples
    save_dir =  config.get('BACK-UP','save_dir')

    # training parameters
    epsilon = config.getfloat('OPTIMIZER_PARAMETERS','epsilon') # overall learning rate for stochastic gradient descent
    epsilonVF = 1.*epsilon # learning rate for VF matrix
    epsilonFH = 0.01*epsilon # learning rate for FH matrix
    epsilonb = 0.01*epsilon # learning rate for scaling parameter coeff
    epsilonw_mean = 0.2*epsilon # learning rate for mean parameters w_mean
    epsilonb_mean = 0.1*epsilon # learning rate for bias_mean
    weightcost_final =  config.getfloat('OPTIMIZER_PARAMETERS','weightcost_final') # weight decay

    # HMC setting
    hmc_step_nr = config.getint('HMC_PARAMETERS','hmc_step_nr') # number of leap-frog steps
    hmc_step =  0.01 # initial step size
    hmc_target_ave_rej =  config.getfloat('HMC_PARAMETERS','hmc_target_ave_rej') # target rejection rate of HMC samples
    hmc_ave_rej =  hmc_target_ave_rej # initial value of running average of rejection rate

    # initialize parameters
    VF = 0.02 * gpu.randn(num_vis, num_fac) # VxH
    FH = gpu.eye(num_hid_cov) # HxO
    coeff = gpu.ones((num_hid_cov, 1))
    bias_vis = gpu.zeros((num_vis, 1))
    w_mean = 0.05 * gpu.randn(num_vis, num_hid_mean) # VxH 
    bias_mean = -1.0* gpu.ones((num_hid_mean,1))
    # auxiliary parameters for Fast PCD, initialized at 0
    VFf = gpu.zeros((num_vis, num_fac)) # VxH
    VFt = gpu.zeros((num_vis, num_fac)) # VxH
    w_meanf = gpu.zeros((num_vis, num_hid_mean)) # VxH
    bias_meanf = gpu.zeros((num_hid_mean,1))
    w_meant = gpu.zeros((num_vis, num_hid_mean)) # VxH
    bias_meant = gpu.zeros((num_hid_mean,1))

    # initialize variables to store derivatives 
    VFinc = gpu.zeros((num_vis, num_fac))
    FHinc = gpu.zeros((num_fac, num_hid_cov))
    coeffinc = gpu.zeros((num_hid_cov, 1))
    bias_visinc = gpu.zeros((num_vis, 1))
    w_meaninc = gpu.zeros((num_vis, num_hid_mean))
    bias_meaninc = gpu.zeros((num_hid_mean, 1))

    # initialize temporary storage variables
    data = gpu.empty((num_vis, batch_size)) # VxP
    negdataini = gpu.empty((num_vis, batch_size)) # VxP
    feat = gpu.empty((num_fac, batch_size))
    featsq = gpu.empty((num_fac, batch_size))
    negdata = gpu.randn(num_vis, batch_size)
    old_energy = gpu.zeros((1, batch_size))
    new_energy = gpu.zeros((1, batch_size))
    gradient = gpu.empty((num_vis, batch_size)) # VxP
    thresh = gpu.zeros((1, batch_size))
    feat_mean = gpu.empty((num_hid_mean, batch_size))
    vel = gpu.randn(num_vis, batch_size)
    normVF = 1    
    
    # other temporary vars
    t1 = gpu.empty((num_hid_cov, batch_size))
    t2 = gpu.empty((num_hid_cov, batch_size))
    t4 = gpu.empty((1,batch_size))
    t5 = gpu.empty((1,1))
    t12 = gpu.empty((num_hid_cov, 1))
    tt0 = 15.*gpu.ones((num_hid_mean, batch_size))
    tt1 = gpu.empty((num_hid_mean, batch_size))
    tt2 = gpu.empty((num_hid_mean, batch_size))

    # start training
    for epoch in range(num_epochs): # loop over all training samples

        print "Epoch " + str(epoch + 1)
    
        # anneal current learning rates
        epsilonVFc    = epsilonVF/max(1.,epoch/10.)
        epsilonFHc    = epsilonFH/max(1.,epoch/10.)
        epsilonbc    = epsilonb/max(1.,epoch/10.)
        epsilonw_meanc = epsilonw_mean/max(1.,epoch/10.)
        epsilonb_meanc = epsilonb_mean/max(1.,epoch/10.)
        weightcost = weightcost_final

        if epoch <= startFH:
            epsilonFHc = 0 
        if epoch <= startwd:	
            weightcost = 0

        for batch in range(num_batches): # loop over batches
             
            # get current minibatch
            data = dev_dat[:,batch*batch_size:(batch + 1)*batch_size] # VxP (nr dims x nr samples)
            
            
            feat = gpu.dot(VF.T, data)
            featsq = feat**2
            t1 = 1 + 0.5 * gpu.dot(FH.T, featsq)
            t2 = coeff / t1
            
            feat_mean = - (gpu.dot(w_mean.T, data) + bias_mean).logistic()
            
            # positive sample increments
            FHinc = gpu.dot(featsq, t2.T)
            VFinc = gpu.dot(data, (gpu.dot(FH, t2) * feat).T)
            coeffinc = (t1.log()).sum(axis=1)
            bias_visinc = - data.sum(axis=1)

            w_meaninc = gpu.dot(data, feat_mean.T)
            bias_meaninc = feat_mean.sum(axis = 1)

            # update total weight vector
            if doFPCD == 1:
                VFt = VF + VFf
                w_meant = w_mean + w_meanf
                bias_meant = bias_mean + bias_meanf
            else:
                VFt = VF
                w_meant = w_mean
                bias_meant = bias_mean
            
            # HMC sampling: draw an approximate sample from the model (run for only one MC step)
            if doFPCD == 1:
		negdataini = negdata
	    else:
		negdataini = data
            negdata, hmc_step, hmc_ave_rej = draw_HMC_samples(negdataini,negdata,vel,gradient,new_energy,old_energy,VFt,FH,coeff,bias_vis,w_meant,bias_meant,hmc_step,hmc_step_nr,hmc_ave_rej,hmc_target_ave_rej,t1,t2,t4,t5,tt0,tt1,tt2,thresh,feat,featsq,batch_size,feat_mean)
            

            feat = gpu.dot(VFt.T, negdata)
            featsq = feat**2
            t1 = 1 + 0.5 * gpu.dot(FH.T, featsq)
            t2 = coeff / t1

            feat_mean = (gpu.dot(w_meant.T, negdata) + bias_meant).logistic()

            # negative sample increments
            FHinc -= gpu.dot(featsq, t2.T)
            VFinc -= gpu.dot(negdata, (gpu.dot(FH, t2) * feat).T)
            coeffinc -= (t1.log()).sum(axis=1)
            bias_visinc += negdata.sum(axis=1)

            w_meaninc += gpu.dot(negdata, feat_mean.T)
            bias_meaninc += feat_mean.sum(axis = 1)

            
            # update parameters
            if doFPCD == 1:
                # fast parameters
                VFf = 19./20 * VFf - epsilonVF/batch_size * VFinc
                w_meanf = 19./20 * w_meanf - epsilonw_mean/batch_size * w_meaninc
                bias_meanf = 19./20 * bias_meanf - epsilonb_mean/batch_size * bias_meaninc.reshape((num_hid_mean,1))
            # regular parameters
            VF -= epsilonVFc/batch_size * (weightcost * VF + VFinc)
            # normalize columns of VF: normalize by running average of their norm 
            norms = gpu.sqrt((VF**2).sum(axis = 0))
            avg_norm = norms.sum()
            normVF = .95*normVF + (.05/num_fac) * avg_norm
            VF = (VF/norms)*normVF

            coeff -= epsilonbc/batch_size * coeffinc.reshape((num_hid_cov,1))
            # make sure coeff does not go negative (actually below .1)
            t12 = coeff > .1
            coeff *= t12
            coeff += (-.1 * t12 + .1)
            # average scale parameters
            coeff = gpu.ones((num_hid_cov, 1)) * coeff.sum() / num_hid_cov
            # update visible biases
            bias_vis -= epsilonbc/batch_size * bias_visinc.reshape((num_vis,1))

            if epoch > startFH:
                FH -= epsilonFHc/batch_size * (FHinc + weightcost * FH.sign())
	        # set to 0 negative entries in FH
                FH *= (FH > 0)         
                if apply_mask==1:
                    FH *= mask # this is useful to enforce a topography
		# normalize columns of FH: L1 norm set to 1 in each column
                FH = FH / FH.sum(axis=0)
            # update mean parameters
            w_mean -= epsilonw_meanc/batch_size * (w_meaninc + weightcost * w_mean)
            bias_mean -= epsilonb_meanc/batch_size * bias_meaninc.reshape((num_hid_mean, 1))

        if verbose == 1: # print some info
            print "VF: " + '%3.2e' % VF.euclid_norm() + ", DVF: " + '%3.2e' % (VFinc.euclid_norm()*(epsilonVFc/batch_size)) + ", FH: " + '%3.2e' % FH.euclid_norm() + ", DFH: " + '%3.2e' % (FHinc.euclid_norm()*(epsilonFHc/batch_size)) + ", coeff: " + '%3.2e' % coeff.euclid_norm() + ", Dcoeff: " + '%3.2e' % (coeffinc.euclid_norm()*(epsilonbc/batch_size)) + ", bias_vis: " + '%3.2e' % bias_vis.euclid_norm() + ", Dbias_vis: " + '%3.2e' % (bias_visinc.euclid_norm()*(epsilonbc/batch_size)) + ", wm: " + '%3.2e' % w_mean.euclid_norm() + ", Dwm: " + '%3.2e' % (w_meaninc.euclid_norm()*(epsilonw_meanc/batch_size)) + ", bm: " + '%3.2e' % bias_mean.euclid_norm() + ", Dbm: " + '%3.2e' % (bias_meaninc.euclid_norm()*(epsilonb_meanc/batch_size)) + ", step: " + '%3.2e' % hmc_step  +  ", rej: " + '%3.2e' % hmc_ave_rej 
            sys.stdout.flush()
        # back-up every once in a while 
        if np.mod(epoch,10) == 0:
            savemat(save_dir + "params_temp", {'VF':VF.as_numpy_array() ,'FH':FH.as_numpy_array(),'coeff': coeff.as_numpy_array(), 'bias_vis': bias_vis.as_numpy_array(),'w_mean': w_mean.as_numpy_array(), 'bias_mean': bias_mean.as_numpy_array(),'data':data.as_numpy_array(),'negdata':negdata.as_numpy_array(),'epoch':epoch})    
    # final back-up
    savemat(save_dir + "params_fac" + str(num_fac) + "_cov" + str(num_hid_cov) + "_mean" + str(num_hid_mean), {'VF':VF.as_numpy_array(),'FH':FH.as_numpy_array(),'coeff': coeff.as_numpy_array(), 'bias_vis': bias_vis.as_numpy_array(), 'w_mean': w_mean.as_numpy_array(), 'bias_mean': bias_mean.as_numpy_array(), 'epoch':epoch})



if __name__ == "__main__":
  train_mHPoT()


